##################### PACKAGE Trimble_DL #####################
#
# (Ver 1.999999999) Dave Mencin UNAVCO dmencin@unavco.ucar.edu
# This is the interface to be called directly and contains no direct dependence on rfile
# Expected tokens:
# ($return, %sitehash ) = get_sitefile( $filename ) returns a $result and 
# hash of site variables keyed by index.
# $result = get_allfies( $filename ) returns success or failure.
# Success means that either maximum attempts were made OR a successful
# download happened. (i.e. program was successful).
#
#
# Alpha Creation: Apr-2 1997
# Added Comments and dummy $result return: Apr-7 1997
# Fixed $return values: Apr-15 1997
# Added scale_factor: Apr-30 1997
# Fixed scoping bug (used my inside an if statement): May-2 1997
# Added TRY_LOOP: May-3 1997
# Added defintion of SUCCESS: May-7 1997
# Added uname support and logic to not dwnld current file: May-9 1997
# Added logic for session.f_retry a second loop for files: May-13 1997
# Added logic for session.p_retry a third loop for purges: May-13 1997
# Revisit and check ldm logic: May-15 1997
# Add seperate Log file directory: May-23 1997
# Add undocumented logic for FREEWAVE: May-28 1997
# Changed filename again (removed 24 and made four digit year: May-28 1997
# Fixed time stamp on log files (TVH): Jun-18,1997
# Fixed warning from overwrite of scope of $nf: Jul-16 1997 
# Fixed non-fatal error ($imon was adding an extra month): Jul-29 1997
# Fixed time_hash error in "biggest" logic: July-29 1997
# Fixed logic that would not hang radio-modem with one file: Jul-29 1997
# Fixed non-fatal error with $rp setting: Oct-9 1997
# Added tml_send_file routine (internal only): Oct-10 1997
# Added logic to support tml if "transport.tml : YES" is in SITE file: Oct-10 1997
# Added full POSIX compliance and use SERIAL.pm version 2.0: Jan-07-1999
# Added DTR drop for FREEWAVE: Jan-07-1999
# Added support for get_serial "get.serail": Jan-07-1999
# Added support for short filename "test.shortname": Jan-07-1999
# Added BAR filename support with daily index: Apr-06-1999
# Fixed known Y2K problems (leap year): Apr-06-1999
# Fixed TRY_LOOP logic error (dtr would not drop if download failed): Jul-29-1999
# Changed termination message to avoid confusion of "success": Jul-29-1999
# Changed "get.serial" will now only print serial number to Log file: Aug-03-1999
# Added "get.serial.filename" will add serial number to filename(only with get.serial): Aug-03-1999
# Added "get.4char" will get the 4 char ID for the receiver instead of site file: Aug-03-1999
# Added "BARGEN.filename" support as indicated by Jim N.: Aug-25-1999
# Fixed BARGEN filenames with zero padding (thanks to Braunj): Sep-29-1999
# Zero padded the log file names so they sorted right: Sep-30-1999
# Added a FREEWAVE option that does not drop DTR, FREEWAVE_NODTR: Nov-11-1999
# Added a very stupid modem dialer (man_dial): Nov-11-1999 
# Added subdirs in the Logs dir for each site: Nov-16-1999
# Fixed logic error, changed && to || in FREEWAVE-MODEM logic: Jan-18-2000
# Fixed error in Leap Year calculation.  Error still exists, but wont go into effect unil 2100:  Jan 21, 2000 
# Fixed error in Leap Year Calculation in two places,for $iy and $cy: Feb. 17, 2000.
# Took out a "_" in the test.shortname filenaming routine: Feb. 17, 2000.
# Removed BAR.filename option, use BARGEN.filename instead: Feb 22, 2000.   
# Fixed error in Leap Year Calculation, for $iy:  March 2, 2000. 
# Normandeau added rsurvey_comm Nov 15 2002
# Added Trimble R7 funtionality : May 23 2005
# Added support for IP connections : 29 APR 08 - JEN


package Trimble_DL;
@ISA = qw(Exporter);
@EXPORT = qw( get_sitefile get_allfiles check_file);

use Trimble_RFILE;
use SERIAL;
use ERROR;
use POSIX;



#################################################################
sub check_file {
	#stub for qc etc.
	return 0;
}
#################################################################
sub tml_send_file {
	#stub for tml
	my $name = shift;
	
	# issue tml  command here ....

	my $tml_command = "tml " . $name;
	error_m("Using $tml_command for transport mechanism.");

	my $ret_code = 0xffff & system( $tml_command );

	if( $ret_code == 0 ) {
		error_m("tml succesful return code detected.");
		return 0;
	} elsif ( $ret_code == 0x00ff ) {
		error_w("tml failed to transport.");
		return 0;
	} elsif ( $ret_code > 0x80 ) {
		$ret_code >>= 8;
		error_w("Non-zero return code from tml of $ret_code.");
		return 0;
	} else {
		error_w("Unknown problem and possible coredump from tml");
		return 0;
	}

	return 0;
}
#################################################################
 sub get_sitefile {

 #Given a filename return a hash of the key/values for the site

 my $filename = shift;
 my %site_info;

 open( SITE, "<" . $filename) or die "Error opening site file\n";

   while(<SITE>){
	next if /#/;
	chomp;
	if (/:/) {
		($key, $value) = split(/:/);
		$key =~ s/^\s*(.*?)\s*$/$1/; # Strip the whitespaces
		$value =~ s/^\s*(.*?)\s*$/$1/;
		$site_info{$key} .= $value;
	}
   }
 return %site_info;
 }
#################################################################
# returns 0 upon sucess and 1 on fail

sub get_allfiles {
	my $SUCCESS = 0;
	my $FAIL = 1;

	my $SITE = shift;


	# Create Log File here ......
	# Stamp it with the unix time first

	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = 
		gmtime(time);
#  For normal date with month from 1 to 12, add 1 to $mon -- TVH Jun 18,1997
        $mon+=1;
	if( $year < 80 ) { $year += 2000; } else { $year += 1900; }
	if( !(-d "Logs") ) { 
		print STDOUT "Creating the sub dir Logs.\n";
		if(!mkdir("Logs",0777)) {
			die "Exiting could not create ./Logs in executable directory\n" ; 
		}
	}
	#See if there is a proper subdirectory, if not create it
	my $log_filename = "BLAH";

	if( !(-d "Logs/$SITE") ) {
		print STDOUT "Trying to create the sub dir ./Logs/$SITE.\n";
		if(!mkdir("Logs/$SITE",0777)) {
			print STDOUT "Could not create ./Logs/$SITE, using ./Log.\n";
			$log_filename = ( "Logs/$SITE." . sprintf("%04d",$year) . "-" 
			. sprintf("%02d",$mon) . "-" . sprintf("%02d",$mday) . "." .
			sprintf("%02d",$hour) . ":" . sprintf("%02d",$min) . ":" .
			sprintf("%02d",$sec) . ".log" );
		} else {
                        $log_filename = ( "Logs/$SITE/$SITE." . sprintf("%04d",$year) . "-"
                        . sprintf("%02d",$mon) . "-" . sprintf("%02d",$mday) . "." .
                        sprintf("%02d",$hour) . ":" . sprintf("%02d",$min) . ":" .
                        sprintf("%02d",$sec) . ".log" );
		}
	} else {
                $log_filename = ( "Logs/$SITE/$SITE." . sprintf("%04d",$year) . "-"
                . sprintf("%02d",$mon) . "-" . sprintf("%02d",$mday) . "." .
                sprintf("%02d",$hour) . ":" . sprintf("%02d",$min) . ":" .
                sprintf("%02d",$sec) . ".log" );
	}

	new_log($log_filename);

	if ( open(UN, "uname -a |") ) {
		select(UN); $| = 1;
		while(<UN>) {
			chomp $_;
			error_i("System: $_");
		}
		close(UN);
	} else {
		error_w("Unable to obtain System information.");
	}

	#Expect the SITE name as the argument
	# Read in the Configuration file
	error_m("Attempting to get configuration file for $SITE");
	%site = get_sitefile( "Sites/". $SITE . ".site" );


	# Print out the variables that exist - may pin point potential problems
	foreach $key (sort keys(%site)) {
		error_m( "$SITE:$key -> $site{$key}");
	}

	# Test Directory info

	if( !(-d $site{'env.downloadpath'})) { error_f("Unable to locate Download directory"); }
	if( $site{'port.connection'} eq "MODEM" ) {
		if( !(-d $site{'env.modempath'})) { error_f("Unable to locate Modem directory"); }
	}

	# Test for exist of rfile

	if( !(-x ($site{'env.rfilepath'} . "rfile")))
		 { error_f("Problem with rfile: either no permission or nonexistent"); } 
	
	local $num_tries = 1;
	TRY_LOOP: while( $num_tries <= $site{'session.retry'} ) {

	error_m("Making attempt number $num_tries of " . $site{'session.retry'}); 
	
	#determine parity for rfile
	my $rp = "e"; # This is the default

	$rp = "n" if $site{'port.control'} =~ "N";
	$rp = "o" if $site{'port.control'} =~ "O";
	$rp = "e" if $site{'port.control'} =~ "E";

	error_m("Attempting to get file listing for $SITE");
	#Build site array
#	my @site = ( $site{'env.rfilepath'}, $site{'port.port'},
		#$site{'port.speed'}, $rp);
#  fix for IP downloads
my @site;
if ("$site{'port.ipport'}") {
	 @site = ( $site{'env.rfilepath'}, "$site{'port.port'}:$site{'port.ipport'}",
		$site{'port.speed'}, $rp);
	error_m("USING PARAMETERS: $site{'port.port'}:$site{'port.ipport'} $site{'port.speed'} $rp  ");
}else {
	 @site = ( $site{'env.rfilepath'}, $site{'port.port'}, 
		$site{'port.speed'}, $rp);
	error_m("USING PARAMETERS: $site{'port.port'} $site{'port.speed'} $rp  ");
}
	my $result;
	my %site_fl;
	my $mcf;


	if( $site{'port.connection'} eq "MODEM" || $site{'port.connection'} eq "MODEM-FREEWAVE" ) {
		
		if ( -e ( $site{'env.modempath'} . $site{'modem.type'}) ) {
			error_m("Found a modem config file for" .
				" $site{'modem.type'}");
			$mcf = $site{'env.modempath'} . $site{'modem.type'};
		} else {
			error_m("Could not find modem config file for" .
				" $site{'modem.type'}");
			if( -e ( $site{'env.modempath'} . "generic")) {
				error_m("Using Generic Modem.");	
				$mcf = $site{'env.modempath'} . "generic";
			} else {
				my $temp = $site{'env.modempath'} .
				$site{'modem.type'};
				error_w("Could not find $temp.");
				error_w("Could not find replacement.");
				error_f("No modem config files.");
				return $FAIL;
			}
		}
	
	error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );	

	$result = connect( @site, $site{'modem.dialstring'} , $mcf );
	if ( error_trimble($result) ) { $num_tries++; next TRY_LOOP; }

	}

	if( $site{'port.connection'} eq "FREEWAVE") {

		error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );

		my $fd = open_serial_port( $site{'port.port'},
			$site{'port.speed'},$site{'port.control'});
		my $bytes = POSIX::write($fd, $site{'modem.dialstring'},
			length($site{'modem.dialstring'}) );
	}

	if( $site{'port.connection'} eq "FREEWAVE_NODTR") {

                my $fd = open_serial_port( $site{'port.port'},
                        $site{'port.speed'},$site{'port.control'});
                my $bytes = POSIX::write($fd, $site{'modem.dialstring'},
                        length($site{'modem.dialstring'}) );
        }

	if( $site{'port.connection'} eq "MODEM-FREEWAVE") {
		my $bytes = POSIX::write($fd, $site{'modem.dialstring'},
			length($site{'modem.dialstring'}) );
	} # There is no way to error check this at present
	
	# Find out what type of Rx is connected
	($result,$model) = get_model(@site,"DIRECT");
	# for output file naming
	#if($model ne "5700") {$ext = ".r00";}
	#else {$ext = ".t00"};
	if ($model eq "R7") {$ext = ".t01";}
	elsif ($model eq "5700") {$ext = ".t00";}
	else {$ext = ".r00";}
	
	error_i("Receiver type is $model");
print STDOUT "###-> STDOUT Receiver type is ->${model}<-\n";
	error_trimble($result); 
# go to sleep here to handle network latency issues

print STDOUT "###-> STDOUT Sleeping 2 seconds.\n";
sleep 2;
	
	# Now get every file on the list *we should be connected in any case *
print STDOUT "###-> STDOUT getting file list.\n";
	($result, %site_fl) = get_filelist(@site,
		 "DIRECT",$site{'modem.scale_factor'},$model);
	if ( error_trimble($result) ) { 
		$num_tries++;
		if( $site{'port.connection'} eq "MODEM") {
			$result = disconnect( @site, $mcf );
			# This should tell us if the modem was the problem ...
			error_trimble($result);
		}
		next TRY_LOOP;
	}


	if ( $site{'get.serial'} eq "YES" ) {
		($result,$serial) = get_serial(@site,"DIRECT");
		error_i("Receiver Serial Number is $serial");
		error_trimble($result);
print STDOUT "##-> STDOUT get serial $result,$serial\n";
	}
	
	

	# Remove the newest file from the list in case it is *current*

	my %time_hash;

	foreach $i ( keys %site_fl ) {

		my ($ih,$im,$is) = split /:/, $site_fl{$i}{'time'};
		my $secs = $ih * 3600 + $im * 60 + $is;
		
		#if($model ne "5700") {
		if(($model ne "R7") and ($model ne "5700")) {

		 	($iy,$imon,$id) = split( /\./, $site_fl{$i}{'date'});
				if( $iy < 80 ) { $iy += 2000; } else { $iy += 1900; }
print STDOUT "###->STDOUT data of non R7 file $imon,$id,$iy\n";
		} else {
		 		($imon,$id,$iy) = split( /-/, $site_fl{$i}{'date'}); # for 5700???
print STDOUT "###->STDOUT data of  R7 file $imon,$id,$iy\n";
		}
		
	
		
		
		if ( (($iy % 4 == 0) && ($iy % 100 != 0)) || ($iy % 400 == 0) ) {
			@mon = (31,29,31,30,31,30,31,31,30,31,30,31);
		} else {
			@mon = (31,28,31,30,31,30,31,31,30,31,30,31);
		}

		# DOY 
		my $doy;
		for $dt (0..($imon-2)) {  $doy += $mon[$dt]; }
		$doy += $id;

		$time_hash{$i} = [ $iy, $doy, $secs ];
	
	}

	my $YEAR = 0; my $DOY = 0; my $SECS = 0; my $biggest = 0;

	foreach $i ( keys %time_hash ) {

	if( $time_hash{$i}[0] > $YEAR ) {
		$DOY = $time_hash{$i}[1];
		$SECS = $time_hash{$i}[2];
		$YEAR = $time_hash{$i}[0];
		$biggest = $i;
	}

       if( $time_hash{$i}[0] == $YEAR ) {
                if( $time_hash{$i}[1] > $DOY ) {
                        $DOY = $time_hash{$i}[1];
                        $SECS = $time_hash{$i}[2];
                        $YEAR = $time_hash{$i}[0];
                        $biggest = $i;
                }
        }

       if( $time_hash{$i}[0] == $YEAR ) {
                if( $time_hash{$i}[1] == $DOY ) {
                        if( $time_hash{$i}[2] > $SECS ) {
                                $DOY = $time_hash{$i}[1];
                                $SECS = $time_hash{$i}[2];
                                $YEAR = $time_hash{$i}[0];
                                $biggest = $i;
                        }
                }
        }

	}

	delete $site_fl{$biggest};
	error_m("Removing index $biggest from current file list.");

	# Did we remove the only file ??
	$nf = keys(%site_fl);
print STDOUT "###->STDOUT  There are $nf files on the receiver\n";
	error_i("There are $nf available file(s)");
	if($nf == 0) { error_m("No files to download"); 

print STDOUT "###->STDOUT  ERROR There are $nf files on the receiver\n";
# Need to issue a rsurvey command 
                error_m("Issuing RSURVEY command to restart rx");
                rsurvey_comm(@site,$SITE,$model);
        }

	foreach $index ( keys %site_fl ) {

		error_m("Getting file index $index");
		my $filename = $site{'env.downloadpath'} . $site_fl{$index}{'name'};
		error_m("Placing data in temp $filename");

		my $file_attempts = 0;

		FILE_LOOP: while( $file_attempts <= $site{'session.f_retry'}){

			error_m("Trying download $file_attempts of " .
				$site{'session.f_retry'}); 
			
			$result = get_file(@site,"DIRECT", $index, $filename,
				$site{'session.fastdownload'},
				$site{'modem.scale_factor'});
			if ( error_trimble($result) ) { 
				if( $file_attempts == $site{'session.f_retry'} ){
					$num_tries++;
					if( $site{'port.connection'} eq "MODEM") {
						$result = disconnect( @site, $mcf );
						error_trimble($result);
					}
					next TRY_LOOP;
				} else {
					$file_attempts++;
					next FILE_LOOP;
				}
			}

			last FILE_LOOP; # Upon Success

		} # end FILE_LOOP

		# Rename file here
		# Use the creation time in a filename format as follows.
		# monumentserial_downloadmachine_YYYY-MM-DD_HHMM[SS.SS].ext
		# where ext is derived from type of data

		my ($ch,$cm,$cs) = split( /:/, $site_fl{$index}{'time'});
	    
		#if($model ne "5700") {
		if(($model ne "R7") and ($model ne "5700")) {

	             ($cy,$cmon,$cd) = split( /\./, $site_fl{$index}{'date'});
				  my $foobar = $cy;	     		  
			  if( $cy < 80 ) { $cy += 2000; } else { $cy += 1900;}
		  } else { 

				   ($cmon,$cd,$cy) = split( /-/, $site_fl{$index}{'date'}); # for 5700???

print STDOUT " \n\n###->STDOUT R7 file date  $cmon,$cd,$cy\n\n\n";
		  }
		
	

		if ( (($cy % 4 == 0) && ($cy % 100 != 0)) || ($cy % 400 == 0) ){
                        @mon = (31,29,31,30,31,30,31,31,30,31,30,31);
                } else {
                        @mon = (31,28,31,30,31,30,31,31,30,31,30,31);
                }

                # DOY
                my $fdoy;
                for $dt (0..($cmon-2)) {  $fdoy += $mon[$dt]; }
                $fdoy += $cd;

		if( $site{'get.4char'} eq "YES" ) {
		  $site4char = $site_fl{$index}{'name'};
		  $site4char =~ s/^(....).*/$1/;
		} else {
		  $site4char = $site{'site.id'};
		}

		my $new_name = ( $site{'env.downloadpath'} . $site4char . "_" . 
		$site{'site.monument_id'} . "_" . $site{'site.computer_id'}
		 . "_" . $cy . "-" . $cmon . "-" . $cd . "_" . $ch .
		 $cm . $cs . $ext ); 
		
print STDOUT " \n\nSTDOUT new name will be $new_name\n\n\n";
		if( $site{'test.shortname'} eq "YES" ) {
			$new_name =  $site{'env.downloadpath'} . $site4char .
			"_" . $cy . "-" . $cmon . "-" . $cd . "_" . $ch . ":" .
			$cm . ":" . $cs . $ext;
		}

		if( $site{'get.serial.filename'} eq "YES" && $site{'get.serial'} eq "YES" ) {
			$new_name =  $site{'env.downloadpath'} . $site4char . "_" .
			$serial . "_" . $cy . "-" . $cmon . "-" . $cd . "_" . $ch . ":" .
			$cm . ":" . $cs . $ext;
		}

		if( $site{'BARGEN.filename'} eq "YES" ) {

			my $minutes;

			$minutes = ($ch * 60) + $cm;

			if( $minutes == 1439 ) {
				$minutes = 0;
				$fdoy += 1;
			}

			$new_name = $site{'env.downloadpath'} . $site4char . sprintf("%03d", $fdoy) . sprintf("%04d", $minutes) . $ext;
			

		}	

		error_i("Renaming file to $new_name");
		#if( $model ne "5700"){
		if(( $model ne "R7") and ( $model ne "5700")){
print STDOUT " STDOUT Renaming file not R7 to $new_name\n";
		rename (($filename . $ext) , $new_name );
		} else {
print STDOUT " STDOUT Renaming $filename from  R7 to $new_name\n";
			rename (($filename) , $new_name );}
		

		#check file here (the logic here may change if want to reload)
		error_m("Checking file $new_name");
		if( check_file($new_name) ) { return $FAIL; }

		if( $site{'transport.tml'} eq "YES") {
			 tml_send_file($new_name); 
		}

		#purge file here

		my $purge_attempts = 0;

		PURGE_LOOP: while( $purge_attempts <= $site{'session.p_retry'}){

			error_m("Purging downloaded file");
			$result = purge_file(@site,"DIRECT",$index);
			if ( error_trimble($result) ) { 
			if( $purge_attempts == $site{'session.p_retry'} ){
				$num_tries++;
				if( $site{'port.connection'} eq "MODEM") {
						$result = disconnect( @site, $mcf );
						error_trimble($result);
					}
					next TRY_LOOP;
				} else {
				$purge_attempts++;
					next PURGE_LOOP;
				}
			}

			last PURGE_LOOP; # Upon Success

		} # end PURGE_LOOP

	} # end FOR_EACH

	if( $site{'port.connection'} eq "MODEM" ||  $site{'port.connection'} eq "MODEM-FREEWAVE") {
		# Disconnect the MODEM	
		$result = disconnect( @site, $mcf );
		error_trimble($result); 

	}

	if( $site{'port.connection'} eq "FREEWAVE") {
		# Disconnect the FREEWAVE
		error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );
		error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );

	}

	 if( $site{'port.connection'} eq "FREEWAVE_NODTR") {
                # Disconnect the FREEWAVE
                error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );
                error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );

        }

	last TRY_LOOP;

	} # end TRY_LOOP

	if( $site{'port.connection'} eq "FREEWAVE") {
                # Disconnect the FREEWAVE
                error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );
                error_w("DTR drop fail") unless drop_dtr( $site{'port.port'} );

        }

	if ( $site{'port.connection'} eq "MODEM-FREEWAVE") {
		#This may generate a bogus error
		$result = disconnect( @site, $mcf );
                error_trimble($result);
	}

	error_m("Program $0 terminated normally, any errors were logged to $log_filename .");
	close_log();

	return $SUCCESS; # SUCCES either means that maximum attempts were made
			 # or a complete download happened !!

}
